unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, ComCtrls,shellapi,MMSystem,
  RiffRWInt,inifiles,AudioIO,deAudioStreams,decommon,deACM,mailerthread,ftpthread,
  ScktComp, ScktComp2,cyclicBuffer,launchthreads,generic,getpwdform;

const
  WM_NOTIFYICON  = WM_USER+333;

type

   typesample = smallint;  // a PCM sample is a 16 bit unsigned int
   psample = ^typesample;

  TfrmMain = class(TForm)
    Panel2: TPanel;
    Panel1: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    ClientSocket1: TClientSocket;
    Panel4: TPanel;
    image1: TImage;
    SpeedButton1: TSpeedButton;
    ServerSocket1: TServerSocket;
    lblMaxRecorded: TLabel;
    Memo1: TMemo;
    ProgressBar1: TProgressBar;
    lblTriggerVolume: TLabel;
    Label3: TLabel;
    TrackBar1: TTrackBar;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormDestroy(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure ServerSocket1ClientRead(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ServerSocket1Accept(Sender: TObject; Socket: TCustomWinSocket);
    procedure ServerSocket1ClientDisconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ServerSocket1ClientError(Sender: TObject;
      Socket: TCustomWinSocket; ErrorEvent: TErrorEvent;
      var ErrorCode: Integer);
    procedure ServerSocket1ClientConnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ClientSocket1Disconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ClientSocket1Error(Sender: TObject; Socket: TCustomWinSocket;
      ErrorEvent: TErrorEvent; var ErrorCode: Integer);
    procedure ClientSocket1Connect(Sender: TObject; Socket: TCustomWinSocket);
    procedure TrackBar1Change(Sender: TObject);
  private
    { Private declarations }
    playSound: boolean;
    source: string;
    MailerThread: TMailer;
    FTPThread: TFTPUploader;
    AudioIn: TAudioIn;
    AudioOut: TAudioOut;
    RiffWrite: IRiffRW;
    DataChunkIndex: integer;
    audioOutputBuffer: TCyclicBuffer;
    buffersRead,recordings: integer;
    recording : boolean;
    startedrecording,recordingTime: dword;
    launchOnStartRecord,launchOnEndRecord, workingDirectory: string;
    saveToDisk, ConvertWavToMP3, sendviaemail, sendviaftp, sendtoreceiver: boolean;
    stream : TWinSocketStream;
    socketHandlerThread: TSocketHandlerThread;
    procedure doBufferProcessing(Buffer: pchar; var Size: integer; maxThisBuffer: integer);
    function getMP3Format: TdeWaveFormat;
    procedure ConvertWavFileToMP3(Filename: string);
    function findMaxSample16Bit(Buffer: PChar; Size: Integer): integer;
    procedure drawGraph(image: TImage; x: integer; maxThisBuffer: integer);
    procedure log(msg: string);
  public
    { Public declarations }
    TrayIcon: TNotifyIconData;
    HMainIcon: HICON;
    function OnBufferFilled(Buffer: pchar;var Size: integer): boolean;
    function OnFillBuffer(Buffer: PChar; var Size: Integer): Boolean;
    procedure ClickTrayIcon(var msg: TMessage); message WM_NOTIFYICON;
    Procedure MinimizeClick(Sender:TObject);
    procedure handleReceivedBuffer(buffer: pchar; size: integer);
  end;

const
   appVersion = '1.6';
   confFile = 'soundcapture.conf';
   maxPossibleVolume = 32767; // for 8 bit change this to 255

var
  frmMain: TfrmMain;
  running: boolean;


implementation

{$R *.DFM}

{left-pads a string}
function LPad(s: String; nLength: integer): string ;
begin
     while length(s) < nLength do
           s := '0' + s ;
     result := s ;
end ;


{called when the microphone buffer has some bytes it recorded}
function TfrmMain.OnBufferFilled(Buffer: pchar;var Size: integer): boolean;
begin
   Result := true;
   handleReceivedBuffer(buffer, size);
end;


procedure TfrmMain.handleReceivedBuffer(buffer: pchar; size: integer);
var
   maxThisBuffer: integer;
begin
   // find the max sample in this buffer
   maxThisBuffer := findMaxSample16Bit(buffer, size);
   lblMaxRecorded.caption := inttostr(maxThisBuffer);
   inc(buffersRead);
   drawGraph(image1, buffersRead, maxThisBuffer);
   progressbar1.Position := (maxThisBuffer * 100) div maxPossibleVolume;
   doBufferProcessing(buffer, size, maxThisBuffer);

   if playSound then audioOutputBuffer.pushBuffer(pbyte(buffer), size);
end;


{called when the speaker buffer wants some bytes to play out}
function TfrmMain.OnFillBuffer(Buffer: PChar; var Size: Integer): Boolean;
begin
   result := true;

   if audioOutputBuffer.available < audioOut.BufferSize then
   begin
      application.processmessages;
      sleep(100);
      size := 0;
      exit;
   end;

   size := audioOut.BufferSize;

   // copy bytes from the audio buffer into the pchar buffer used by TAudioOut
   audioOutputBuffer.popBuffer(pbyte(buffer), size);
end;


procedure TfrmMain.doBufferProcessing(Buffer: pchar; var Size: integer; maxThisBuffer: integer);
var
   filename: string;
   elapsedTime: dword;
   f: file;
begin
   // stream to socket if required
   if sendtoreceiver then
   begin
      try
         if not clientsocket1.active then
         begin
            log('Opening socket to ' + clientsocket1.address);
            clientsocket1.open;
            freeandnil(stream);
            stream := TWinSocketStream.Create(ClientSocket1.Socket, 60000);
         end;
         stream.Write(Buffer^, Size)
      except
         on E: Exception do
         clientsocket1.close;
      end;
   end;

   // start recording if threshold reached
   if (saveToDisk) and (not recording) and (maxThisBuffer >= trackbar1.Position) then
   begin
      if length(launchOnStartRecord) > 0 then
      begin
         ShellExecute(self.handle, 'open', pchar(launchOnStartRecord), nil, pchar(workingDirectory), SW_SHOWNORMAL);
         log('Launched ' + launchOnStartRecord);
      end;

      recording := true;
      startedRecording := GetTickCount;
      label2.caption := 'Recording...';
      label2.Font.Color := clRed;
      label2.Font.Style := [fsBold];
      log('Started recording');
    end;

    if recording then
    begin
      // firstly we just update in memory
      with RiffWrite do
      begin
         Chunk[DataChunkIndex].SetSize(Chunk[DataChunkIndex].Size + Size);
         Chunk[DataChunkIndex].Write(Buffer^,Size);
      end;

      // save to file after a certain amount of time
      elapsedTime := GetTickCount - startedRecording;
      if elapsedTime > (recordingTime * 1000) then
      begin

         inc(recordings);
         filename := '_rec_' + lpad(inttostr(recordings), 6) + '.wav';
         label2.caption := 'Saving ' + filename;
         RiffWrite.SaveToFile(filename);

         if ConvertWavToMP3 then
         begin
            ConvertWavFileToMP3(filename);
            deletefile(filename);
            filename := ChangeFileExt(FileName, '.mp3');
         end;

         // as we initially save the file with a _ prefix we need to remove this
         // this is to stop the mailer thread trying to read it as the audio unit is still writing it out
         AssignFile(f, filename);
         rename(f, copy(filename, 2, length(filename)-1));
         log('Saved file ' + filename);

         // reset everything
         RiffWrite.DeleteChunk(DataChunkIndex);
         RiffWrite.AddChunk('data');
         DataChunkIndex := RiffWrite.IndexOf('data');
         label2.caption := 'Monitoring...';
         label2.Font.Color := clBlack;
         label2.Font.Style := [];

         if length(launchOnEndRecord) > 0 then
         begin
            ShellExecute(self.handle, 'open', pchar(launchOnEndRecord), nil, pchar(workingDirectory), SW_SHOWNORMAL);
            log('Launched ' + launchOnEndRecord);
         end;

         recording := false;
         log('Recording ended');
      end;
    end;

    application.processmessages;
end;


procedure TfrmMain.FormCreate(Sender: TObject);
var
  WaveFormat: TWaveFormatEx;
  i,val : integer;
  SRec: TSearchRec;
  IniFile : TIniFile;
  extension: string;
  launchReceiverThread: TLaunchReceiverThread;
  launchSenderThread: TLaunchSenderThread;
  ftpPassword: string;
begin
  running := true;
  self.caption := self.caption + ' ' + appVersion;
  HMainIcon:=LoadIcon(MainInstance, 'MAINICON');

  Shell_NotifyIcon(NIM_DELETE, @TrayIcon);
  with trayIcon do
  begin
    cbSize              := sizeof(TNotifyIconData);
    Wnd                 := handle;
    uID                 := 123;
    uFlags              := NIF_MESSAGE or NIF_ICON or NIF_TIP;
    uCallbackMessage    := WM_NOTIFYICON;
    hIcon               := HMainIcon;
    szTip               := 'Sound Capture';
  end;
  Application.OnMinimize:= MinimizeClick;

  audioOutputBuffer := TCyclicBuffer.create(cyclicBufferSize);

   // read config
   IniFile := TIniFile.Create(ExtractFilePath(Application.ExeName) + '\' + confFile);
   try
      with IniFile do
      begin
         source := ReadString('main', 'source', 'microphone');
         playSound := (ReadString('main', 'playSound', 'false') = 'true');
         trackbar1.position := ReadInteger('main', 'triggerVolume', 20000);
         recordingTime := ReadInteger('main', 'recordingTime', 10);
         launchOnStartRecord := ReadString('main', 'launchOnStartRecord', '');
         launchOnEndRecord := ReadString('main', 'launchOnEndRecord', '');
         workingDirectory := ReadString('main', 'workingDirectory', '');
         saveToDisk := ReadString('main', 'saveToDisk', 'true') = 'true';
         sendViaEmail := ReadString('main', 'sendViaEmail', 'false') = 'true';
         sendViaFTP := ReadString('main', 'sendviaFTP', 'false') = 'true';
         sendToReceiver := ReadString('main', 'sendToReceiver', 'false') = 'true';
         ConvertWavToMP3 := ReadString('main', 'convertWavToMP3', 'true') = 'true';
         clientsocket1.address := ReadString('main', 'soundReceiverHost', '127.0.0.1');
         clientsocket1.port := ReadInteger('main', 'soundReceiverPort', 2001);
         serversocket1.port := ReadInteger('main', 'sourcesocketPort', 2001);
      end;

      // set up disk recording stuff
      if saveToDisk then
      begin
         if ConvertWavToMP3 then
            extension := 'mp3'
         else
            extension := 'wav';

         // get highest numbered audio file in directory
         recordings := 0;
         i := FindFirst(ExtractFilePath(Application.ExeName) + '\rec_*.' + extension, faAnyFile, SRec);
         try
            while i = 0 do
            begin
               val := strtoint(copy(SRec.Name, 5, 6));
               if val > recordings then recordings := val;
               i := FindNext(SRec);
            end;
         finally
            FindClose(SRec);
         end ;

         // if we need to email the sound files, start the background thread
         if sendviaemail then
         begin
            MailerThread := TMailer.create(IniFile, recordings, ConvertWavToMP3);
            MailerThread.FreeOnTerminate := True ;
            MailerThread.resume;
            log('Starting SMTP thread');
         end;

         // if we need to FTP the sound files, start the background thread
         if sendviaftp then
         begin
            ftpPassword := IniFile.ReadString('main', 'ftp.password', '');
            if ftpPassword = '*prompt*' then
               ftpPassword := getPassword('Please enter FTP password');
            FTPThread := TFTPUploader.create(IniFile, recordings, ConvertWavToMP3, ftpPassword);
            FTPThread.FreeOnTerminate := True ;
            FTPThread.resume;
            log('Starting FTP thread');
         end;
      end;

      TrackBar1Change(self);

      image1.canvas.Brush.Color := clBlack;
      image1.canvas.Brush.Style := bsSolid;
      image1.canvas.FillRect(Rect(0,0,image1.width,image1.height));

      caption := caption + ', audio source=' + source;
      if playSound then
         caption := caption + ' (playing sound)'
      else
         caption := caption + ' (not playing sound)';

      if playSound then
      begin
         log('Opening speaker channel');
         AudioOut := TAudioOut.create(self);
         AudioOut.BufferSize := buffersize;
         AudioOut.OnFillBuffer := OnFillBuffer;
         AudioOut.Stereo := stereo;
         AudioOut.quantization := quantization;
         AudioOut.framerate := framerate;
      end;

      if source = 'microphone' then
      begin
         log('Opening microphone');
         AudioIn := TAudioIn.Create(self);
         AudioIn.BufferSize := buffersize;
         AudioIn.OnBufferFilled := OnBufferFilled;
         AudioIn.Stereo := stereo;
         AudioIn.quantization := quantization;
         AudioIn.framerate := framerate;
      end;

      ZeroMemory(@WaveFormat,SizeOf(WaveFormat));
      WaveFormat.wFormatTag := WAVE_FORMAT_PCM;
      WaveFormat.nChannels := byte(stereo)+1;
      WaveFormat.wBitsPerSample := Quantization;
      WaveFormat.nSamplesPerSec := FrameRate;
      WaveFormat.nBlockAlign := WaveFormat.nChannels * (WaveFormat.wBitsPerSample shr 3);
      WaveFormat.nAvgBytesPerSec := WaveFormat.nSamplesPerSec * WaveFormat.nBlockAlign;

      RiffWrite := CreateRiffRW;
      RiffWrite.FourCC := 'WAVE';
      with RiffWrite.AddChunk('fmt ') do Write(WaveFormat,SizeOf(WaveFormat));
      RiffWrite.AddChunk('data');
      DataChunkIndex := RiffWrite.IndexOf('data');

      if AudioIn <> nil then
      begin
         // we need to start AudioIn in a separate thread as it blocks
         launchSenderThread := TLaunchSenderThread.create(AudioIn);
         launchSenderThread.FreeOnTerminate := True ;
         launchSenderThread.resume;
      end;

      if AudioOut <> nil then
      begin
         // we need to start AudioOut in a separate thread as it blocks
         launchReceiverThread := TLaunchReceiverThread.create(AudioOut);
         launchReceiverThread.Priority := tpTimeCritical;
         launchReceiverThread.FreeOnTerminate := True ;
         launchReceiverThread.resume;
      end;

      if source = 'socket' then
      begin
         log('Opening server socket');
         ServerSocket1.Active := true;
         socketHandlerThread := TSocketHandlerThread.create(self);
         socketHandlerThread.Priority := tpTimeCritical;
         socketHandlerThread.FreeOnTerminate := True ;
         socketHandlerThread.resume;
      end;

   finally
   	IniFile.Free ;
   end;
   
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
var
   IniFile: TIniFile ;
begin
   // save settings
   IniFile := TIniFile.Create(ExtractFilePath(Application.ExeName) + '\' + confFile);
   try
      with IniFile do
      begin
         WriteInteger('main', 'triggerVolume',trackbar1.position);
      end;
   finally
   	IniFile.Free ;
   end;

   // stop/free all our objects
   running := false;
   if MailerThread <> nil then MailerThread.terminate;
   if FTPThread <> nil then FTPThread.terminate;
   if (AudioIn <> nil) and (AudioIn.AudioActive) then AudioIn.StopGracefully;
   if (AudioOut <> nil) and (AudioOut.AudioActive) then AudioOut.StopGracefully;

   if clientsocket1.active then clientsocket1.close;
   if serversocket1.active then serversocket1.close;
   audioOutputBuffer.free;
end;

procedure TfrmMain.ServerSocket1Accept(Sender: TObject;
  Socket: TCustomWinSocket);
begin
   log('Accepted incoming connection from ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ServerSocket1ClientConnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
   log('Socket connect from ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ServerSocket1ClientDisconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
   log('Socket disconnect from ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ServerSocket1ClientError(Sender: TObject;
  Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
   log('Socket error from ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ServerSocket1ClientRead(Sender: TObject;
  Socket: TCustomWinSocket);
var
   tmpBuf: array[0..65534] of byte;
   bytesRead: integer;
begin
   bytesRead := Socket.ReceiveBuf(tmpBuf, sizeOf(tmpBuf)-1);
   socketHandlerThread.handleReceivedBuffer(@tmpBuf[0], bytesRead);
end;

procedure TfrmMain.ClientSocket1Connect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
   log('Socket connected ok to ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ClientSocket1Disconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
   log('Socket disconnect from ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ClientSocket1Error(Sender: TObject; Socket: TCustomWinSocket;
  ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
   log('Socket error from ' + Socket.RemoteAddress);
end;

procedure TfrmMain.ConvertWavFileToMP3(Filename: string);
var
  InputStream: TStream;
  WavFile: TFileStream;
  Mp3File: TFileStream;
  WavReader: TdeRIFFAudioReader;
  RawWav: TdeSubStream;
  Mp3Filename: string;
  Converter: TdeAudioConvertor;
  Mp3WaveFormat: TdeWaveFormat;
begin
  Mp3Filename := ChangeFileExt(FileName, '.mp3');
  Mp3File := nil;
  WavReader := nil;
  WavFile := nil;
  try
    WavFile := TFileStream.Create(Filename, fmOpenRead or fmShareDenyWrite);
    Mp3File := TFileStream.Create(Mp3Filename, fmCreate);
    WavReader := TdeRIFFAudioReader.Create(WavFile);
    RawWav := WavReader.GetChunkAsStream(0);
    InputStream := RawWav;
    Converter := TdeAudioConvertor.create(self);
    Converter.FormatIn.Assign(WavReader.Format);
    Mp3WaveFormat := getMP3Format;
    Converter.FormatOut.Assign(Mp3WaveFormat);
    Converter.Activate;
    Converter.OutputStream := MP3File;
    Converter.ConvertAudioStream(InputStream);
  finally
    Converter.free;
    Mp3WaveFormat.free;
    InputStream := nil;
    WavReader.Free;
    WavFile.Free;
    MP3File.Free;
    Converter.Active := False;
  end;
end;


{this function returns the first MP3 codec found on the system}
function TfrmMain.getMP3Format: TdeWaveFormat;
var
  CodecIndex: Integer;
  CurrentCodec: TdeCodec;
  FormatIndex: Integer;
  CurrentFormatInfo: TFormatInfo;
  CurrentFormatTag: Word;
  RegisteredWaveFormat: TdeWaveFormat;
begin
  for CodecIndex := 0 to deACM.CodecCount - 1 do
  begin
    CurrentCodec := deACM.GetCodec(CodecIndex);
    for FormatIndex := 0 to CurrentCodec.SupportedFormatCount - 1 do
    begin
      CurrentFormatInfo := CurrentCodec.SupportedFormat[FormatIndex];
      CurrentFormatTag := CurrentFormatInfo.WaveFormatEx.wFormatTag;
      if CurrentFormatTag = WAVE_FORMAT_MPEGLAYER3 then
      begin
        RegisteredWaveFormat := TdeWaveFormat.Create;
        RegisteredWaveFormat.AssignPWaveFormatEx(CurrentFormatInfo.WaveFormatEx);
        result := RegisteredWaveFormat;
        //ShowMessage(RegisteredWaveFormat.Description);
        exit;
      end;
    end;
  end;
end;


procedure TfrmMain.FormDestroy(Sender: TObject);
begin
   AudioIn.free;
   RiffWrite.clear;
   freeAndNil(stream);
   Shell_NotifyIcon(NIM_Delete, @TrayIcon);
end;

procedure TfrmMain.SpeedButton1Click(Sender: TObject);
begin
   ShellExecute(GetDesktopWindow, 'open', PChar('http://members.lycos.co.uk/wuul/soundcapture/readme.html'), nil, nil, SW_ShowNormal);
end;


procedure TfrmMain.TrackBar1Change(Sender: TObject);
begin
   lblTriggerVolume.caption := inttostr(trackbar1.position);
end;

Procedure TfrmMain.MinimizeClick(Sender:TObject);
begin
   Shell_NotifyIcon(NIM_Add, @TrayIcon);
   hide;
   {hide the taskbar button}
   if IsWindowVisible(Application.Handle)
   then ShowWindow(Application.Handle, SW_HIDE);
end;

procedure TfrmMain.ClickTrayIcon(var msg: TMessage);
begin
  case msg.lparam of
    WM_LBUTTONUP, WM_LBUTTONDBLCLK :
    {WM_BUTTONDOWN may cause next Icon to activate if this icon is deleted -
        (Icons shift left and left neighbor will be under mouse at ButtonUp time)}
    begin
      Application.Restore;  {restore the application}
      If WindowState = wsMinimized then WindowState := wsNormal;  {Reset minimized state}
      {Added 5/6/04 ====>} visible:=true;
      SetForegroundWindow(Application.Handle); {Force form to the foreground }
      Shell_NotifyIcon(NIM_Delete, @TrayIcon);
    end;
  end;
end;


{returns the maximum sample in a buffer - this is used for drawing the graph}
function TfrmMain.findMaxSample16bit(Buffer: pchar; Size: Integer): integer;
var
   b: psample; // a sample is a 16 bit unsigned int
   sample: typesample;
   i, maxThisBuffer: integer;
begin
   maxThisBuffer := 0;
   b := psample(buffer); // cast buffer to a pointer to a "sample" type
   for i := 1 to Size div 2 do // each sample is a 16 bit signed int
   begin
      sample := b^;
      if sample > maxThisBuffer then maxThisBuffer := sample;
      inc(b);
   end;
   result := maxThisBuffer;
end;


{updates one of the graphs to show current volume}
procedure TfrmMain.drawGraph(image: TImage; x: integer; maxThisBuffer: integer);
var
   y: integer;
begin
   if x mod image.width = 0 then
   begin
     image.canvas.pen.color := clBlack;
     image.canvas.FillRect(Rect(0,0,image.width,image.height))
   end
   else if x > image.width then
   begin
      x := x mod image.width;
   end;

   //if bit16Sound then
      y := (maxThisBuffer * image.Height) div maxPossibleVolume; // for 16 bit
   //else
      //y := (maxThisBuffer * image.Height) div (255-outputZeroSoundPressure8bit); // for 8 bit

   if y < 2 then y := 2; // so we always show something, even a straight line

   image.canvas.pen.color := clLime;

   image.Canvas.MoveTo(x, image.Height div 2);
   image.Canvas.LineTo(x, (image.Height div 2) - (Y div 2));

   image.Canvas.MoveTo(x, image.Height div 2);
   image.Canvas.LineTo(x, (image.Height div 2) + (Y div 2));
end;


procedure TfrmMain.log(msg: string);
begin
   if memo1.lines.Count > 100 then memo1.Lines.Clear;
   memo1.Lines.Append(msg);
end;

end.
